
/**************************************************************************
 *
 *  $Id: mbgclock_main.c 1.3.1.1 2017/08/11 10:18:36 martin TEST $
 *
 *  Description:
 *    Main file for for mbgclock driver to support Meinberg bus level
 *    devices under NetBSD.
 *
 *    The binary is a loadable module called mbgclock which implements
 *    /dev/mbgclock* devices.
 *
 *  Copyright (c) 2011-2017 Frank Kardel
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 *  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgclock_main.c $
 *  Revision 1.3.1.1  2017/08/11 10:18:36  martin
 *  Re-added LKM support.
 *  Removed obsolete debug stuff.
 *  Revision 1.3  2017/08/10 16:06:38  martin
 *  String 'pcps_driver_name' is now defined in pcpsdrvr.h.
 *  Account for unified extended resource properties handling.
 *  Account for renamed library symbols.
 *  Revision 1.2  2017/08/10 16:04:28  martin
 *  Updates submitted by Frank Kardel.
 *  Revision 1.1  2011/01/26 13:56:32  martin
 *  Initial revision submitted by Frank Kardel.
 *
 **************************************************************************/

#include <sys/param.h>
#include <sys/lwp.h>
#include <sys/callout.h>
#include <sys/device.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/kauth.h>
#include <dev/pci/pcidevs.h>
#include <dev/pci/pcivar.h>
#include <sys/timetc.h>
#include <sys/timepps.h>
#include <sys/errno.h>
#include <sys/atomic.h>

#include <pcpsdrvr.h>
#include <mbgddmsg.h>

#ifndef PCI_BAR
#define PCI_BAR(__n)            (0x10 + 4 * (__n))
#endif

#define REV_NUM          0x101
#define REV_NUM_STR      "1.01"
#define MBG_COPYRIGHT    "(c) Meinberg 2011, Frank Kardel 2011-2017"
#define MBG_DRVR_NAME    "mbgclock"
#define MBG_VENDOR       "Meinberg Funkuhren"

static MBG_DBG_DATA mbg_dbg_data;
static MBG_DBG_PORT mbg_dbg_port = 0x378 + 0;  //##++
static PCPS_IO_ADDR_MAPPED mbg_dbg_port_mapped;  //##++
static PCPS_DRVR_INFO drvr_info = { REV_NUM, 0, MBG_DRVR_NAME " radio clock driver" };

#include <macioctl.h>

static dev_type_open(mbgclockopen);
static dev_type_close(mbgclockclose);
static dev_type_read(mbgclockread);
static dev_type_write(mbgclockwrite);
static dev_type_ioctl(mbgclockioctl);
static int	mbgclock_match(device_t, cfdata_t, void *);
static void	mbgclock_attach(device_t, device_t, void *);
static int	mbgclock_detach(device_t, int);
static void     mbgclock_tick(void *);
static timecounter_get_t mbgclock_get_timecount;

struct mbgclock_softc
{
	device_t               dev;		/* device backling */
	struct pci_attach_args pa;		/* attach argument */
	int                    isopen;          /* openflag */
	PCPS_DDEV             *pddev;		/* driver handle */
	struct pps_state       pps_state;	/* pps state */
	kmutex_t               mutex;           /* callout termination */
	callout_t              callout;		/* callout structure */
	PCPS_TIME_STATUS_X     last_status;	/* last device status */
	struct timecounter     tc;		/* timecounter - if available  */
};

#define MBG_UNITBITS    6	                /* up to 64 devices */
#define MBG_UNITMASK	((1<<MBG_UNITBITS)-1)
#define MBG_UNIT(x)     (minor(x) & MBG_UNITMASK)
#define MBG_UNITTYPE(x)	(minor(x) >> MBG_UNITBITS)

const struct cdevsw mbgclock_cdevsw = {
	.d_open     = mbgclockopen,
	.d_close    = mbgclockclose,
	.d_read     = mbgclockread,
	.d_write    = mbgclockwrite,
	.d_ioctl    = mbgclockioctl,
	.d_stop     = nostop,
	.d_tty      = notty,
	.d_poll     = nopoll,
	.d_mmap     = nommap,
	.d_kqfilter = nokqfilter,
	.d_flag     = D_OTHER|D_MPSAFE
};

#ifdef CFATTACH_DECL3_NEW
CFATTACH_DECL3_NEW(mbgclock,
		   sizeof(struct mbgclock_softc),
		   mbgclock_match,
		   mbgclock_attach,
		   mbgclock_detach,
		   NULL,
		   NULL,
		   NULL,
		   DVF_DETACH_SHUTDOWN);
#else
CFATTACH_DECL2_NEW(mbgclock,
		   sizeof(struct mbgclock_softc),
		   mbgclock_match,
		   mbgclock_attach,
		   mbgclock_detach,
		   NULL,
		   NULL,
		   NULL);
#endif

extern struct cfdriver mbgclock_cd;

#if defined( DEBUG )
static int mbgclockdebug = ~0;
#define DB_FOLLOW	0x00000001
#define DPRINTF(_X_, _Y_) do {			\
		if (mbgclockdebug & (_X_)) {	\
			printf _Y_;		\
		}				\
	}					\
	while (/*CONST_COND*/0)
#else
#define DPRINTF(_X_, _Y_) /* empty */
#endif

/*----- loadables -----*/
#if defined (_LKM) || defined(_MODULE)

static volatile uint32_t opencount = 0;	/* number of open char drivers */
#define INC_REF atomic_inc_32(&opencount)
#define DEC_REF atomic_dec_32(&opencount)

CFDRIVER_DECL(mbgclock, DV_DULL, NULL);

static int mbgclock_loc[] = { -1, -1 };

static struct cfparent pciparent = {
	"pci", "pci", DVUNIT_ANY
};

static struct cfdata mbgclock_cfdata[] = {
	{
		.cf_name   = "mbgclock",
		.cf_atname = "mbgclock",
		.cf_unit   = 0,
		.cf_fstate = FSTATE_STAR,
		.cf_loc    = mbgclock_loc,
		.cf_flags  = 0,
		.cf_pspec  = &pciparent,
	},
	{ NULL }
};
#else
#define INC_REF
#define DEC_REF
#endif

/*----- module -----*/
#ifdef _MODULE

#include <sys/module.h>

MODULE(MODULE_CLASS_DRIVER, mbgclock, NULL);

static int
mbgclock_modcmd(modcmd_t cmd, void *arg)
{
	int bmajor = -1, cmajor = -1,  error = 0;

	switch (cmd) {
	case MODULE_CMD_INIT:
		DPRINTF(DB_FOLLOW, ("%s: MOD_CMD_INIT\n", mbgclock_cd.cd_name));

		error = config_cfdriver_attach(&mbgclock_cd);
		if (error)
			break;

		error = config_cfattach_attach(mbgclock_cd.cd_name, &mbgclock_ca);
	        if (error) {
			config_cfdriver_detach(&mbgclock_cd);
			aprint_error("%s: unable to register cfattach\n",
			    mbgclock_cd.cd_name);
			break;
		}

		/* s = splaudio(); */
		error = config_cfdata_attach(mbgclock_cfdata, 1);
		/* splx(s); */

		if (error == 0)
		{
			error = devsw_attach("mbgclock", NULL, &bmajor,
					     &mbgclock_cdevsw, &cmajor);
		}

		if (error) {
			config_cfattach_detach(mbgclock_cd.cd_name, &mbgclock_ca);
			config_cfdriver_detach(&mbgclock_cd);
			break;
		} else {
			DPRINTF(DB_FOLLOW, ("%s: registered character major = %d\n",
					    mbgclock_cd.cd_name, cmajor));
		}

		break;

	case MODULE_CMD_FINI:
		DPRINTF(DB_FOLLOW, ("%s: MOD_CMD_FINI\n", mbgclock_cd.cd_name));

		if (opencount) {
			error = EBUSY;
			break;
		}

		error = config_cfdata_detach(mbgclock_cfdata);
		if (error)
			break;

		error = config_cfattach_detach(mbgclock_cd.cd_name, &mbgclock_ca);
		if (error)
			break;

		config_cfdriver_detach(&mbgclock_cd);
		devsw_detach(NULL, &mbgclock_cdevsw);
		break;

	case MODULE_CMD_STAT:
		error = ENOTTY;
		break;

	default:
		error = ENOTTY;
		break;
	}

	DPRINTF(DB_FOLLOW, ("modcmd rc = %d\n", error));

	return error;
}
#endif

/*------ LKM -----*/
#if defined(_LKM) && !defined(_MODULE)

#include <sys/lkm.h>

int mbgclock_lkmentry(struct lkm_table *, int, int);

static struct cfdriver *mbgclock_cfdrivers[] = {
        &mbgclock_cd,
        NULL
};
static struct cfattach *mbgclock_cfattachs[] = {
        &mbgclock_ca,
        NULL
};
static const struct cfattachlkminit mbgclock_cfattachinit[] = {
        { "mbgclock", mbgclock_cfattachs },
        { NULL }
};

MOD_DRV("mbgclock",
        mbgclock_cfdrivers,
        mbgclock_cfattachinit,
        mbgclock_cfdata);

static int
mbgclock_lkmload(struct lkm_table *lkm, int cmd)
{
	int bmajor = -1;
	int cmajor = -1;
	int error;

	error = devsw_attach("mbgclock", NULL, &bmajor,
			&mbgclock_cdevsw, &cmajor);

#ifdef DEBUG
	if (error == 0) {
		DPRINTF(DB_FOLLOW, ("%s: registered character major = %d\n",
				    mbgclock_cd.cd_name, cmajor));
	} else {
		DPRINTF(DB_FOLLOW, ("%s: device driver registry failed errno = %d\n",
				    mbgclock_cd.cd_name, error));
	}
#endif
	return error;
}

static int
mbgclock_lkmunload(struct lkm_table *lkm, int cmd)
{
	if (opencount)
		return EBUSY;

	devsw_detach(NULL, &mbgclock_cdevsw);
	return 0;
}

int
mbgclock_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
{
        DISPATCH(lkmtp, cmd, ver, mbgclock_lkmload, mbgclock_lkmunload, lkm_nofunc);
}
#endif

/*----- autoconf -----*/

/* PCI Support Functions */

static void
mbg_deallocate_resource( EXT_RSRC_INFO *p_ri )
{
	if ( p_ri->valid )
	{
		bus_space_unmap( p_ri->bst, p_ri->bsh, p_ri->size);
		p_ri->valid = 0;
	}
}

/*
 * deallocate resources
 */
static void
mbg_dealloc_rsrcs( struct mbgclock_softc *psc )
{
	PCPS_DDEV *pddev = psc->pddev;

	if ( pddev != NULL )
	{
		PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;
		int i;

		/*  mbg_deallocate_resource( device, &prsrci->irq.bsd, SYS_RES_IRQ ); */

		for ( i = 0; i < N_PCPS_MEM_RSRC; i++ )
			mbg_deallocate_resource( &prsrci->mem[i].ext );

		for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
			mbg_deallocate_resource( &prsrci->port[i].ext );
	}
}

static void
mbg_alloc_rsrc( struct mbgclock_softc *psc , int reg, EXT_RSRC_INFO *p_ri, int type )
{
	p_ri->reg  = reg;
	p_ri->type = type;

	if (!pci_mapreg_map(&psc->pa, reg, type, 0,
			    &p_ri->bst, &p_ri->bsh,
			    &p_ri->base, &p_ri->size))
	{
		p_ri->valid = 1;
	} else {
		p_ri->valid = 0;
	}
}

static void
mbg_alloc_rsrcs( struct mbgclock_softc *psc )
{
  PCPS_DDEV      *pddev      = psc->pddev;
  PCPS_RSRC_INFO *prsrci     = &pddev->rsrc_info;
  EXT_RSRC_INFO   ri;
  pcireg_t	  command;
  int             bar;


  /* clear valid flags */
  for ( bar = 0; bar < prsrci->num_rsrc_io; bar ++)
  	  prsrci->port[bar].ext.valid = 0;

  for ( bar = 0; bar < prsrci->num_rsrc_mem; bar ++)
	  prsrci->mem[bar].ext.valid = 0;

  command = pci_conf_read(psc->pa.pa_pc, psc->pa.pa_tag, PCI_COMMAND_STATUS_REG);

  _mbgddmsg_4( MBG_DBG_INIT_DEV,
		       "%s: PCI device 0x%04X:0x%04X : command status register: 0x%08X",
		       pcps_driver_name, PCI_VENDOR(psc->pa.pa_id),
	       PCI_PRODUCT(psc->pa.pa_id), command );

  for ( bar = 0; bar < 6; bar ++ )
  {
	  int type;
	  int reg = PCI_BAR( bar );

	  _mbgddmsg_4( MBG_DBG_INIT_DEV,
		       "%s: alloc I/O range %i: PCI device 0x%04X:0x%04X",
		       pcps_driver_name, bar, PCI_VENDOR(psc->pa.pa_id),
		       PCI_PRODUCT(psc->pa.pa_id) );

	  switch ((type = pci_mapreg_type(psc->pa.pa_pc, psc->pa.pa_tag, reg)))
	  {
	  case PCI_MAPREG_TYPE_IO:
		  if ( prsrci->num_rsrc_io < N_PCPS_PORT_RSRC )
		  {
			  mbg_alloc_rsrc( psc, reg, &ri, type );

			  if ( ri.valid )
			  {
				  prsrci->port[prsrci->num_rsrc_io].ext = ri;
				  pcps_add_rsrc_io( pddev, ri.base, ri.size );
				  continue;
			  }
			  else
			  {
				  _mbgddmsg_2( MBG_DBG_INIT_DEV,
					       "%s: alloc I/O range %i failed",
					       pcps_driver_name, bar );
			  }
		  }
		  break;

	  case PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_64BIT:
#if __LP64__
#if 0
		  aprint_error_dev(psc->dev,
				   "WARNING: ignoring 64-bit BAR @ 0x%02x\n", reg);
		  continue;
#endif
#else
		/*FALLTHROUGH*/
#endif
	  case PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_32BIT:
		  /* case PCI_MAPREG_TYPE_ROM: currently same as above */
		  if ( prsrci->num_rsrc_mem < N_PCPS_MEM_RSRC )
		  {
			  mbg_alloc_rsrc( psc, reg, &ri, type );

			  if ( ri.valid )
			  {
				  prsrci->mem[prsrci->num_rsrc_mem].ext = ri;
				  pcps_add_rsrc_mem( pddev,
						     (MBG_MEM_ADDR)bus_space_vaddr( ri.bst, ri.bsh),
						     ri.size );
				  continue;
			  }
		  }
		  break;

	  default:
		  printf("XXX unsupported type from pci_mapreg_type: %d\n", type);
		  break;
	  }
  }

  /* single IRQ resource */
#if 0  /* currently not used / required */
  mbg_alloc_rsrc( device, 0, &ri, SYS_RES_IRQ, RF_SHAREABLE | RF_ACTIVE );

  if ( ri.res )
  {
    prsrci->port[prsrci->num_rsrc_irq].ext = ri;
    pcps_add_rsrc_irq( pddev, rman_get_start( ri.res ) );
  }
#endif

}

static int
mbgclock_match(struct device *parent, struct cfdata *match, void *aux)
{
	struct pci_attach_args *pa;
	int16_t vend_id;
	int16_t dev_id;

	DPRINTF(DB_FOLLOW, ("mbgclock_match(0x%p, 0x%p, 0x%p)\n", parent, match, aux));

	pa = aux;
	vend_id = PCI_VENDOR(pa->pa_id);
	dev_id = PCI_PRODUCT(pa->pa_id);

	if (vend_id == PCI_VENDOR_MEINBERG)
	{
		PCPS_DEV_TYPE *pdt;

		pdt = pcps_get_dev_type_table_entry( PCPS_BUS_PCI, dev_id );

		if (pdt != NULL)
		{
			_mbgddmsg_3( MBG_DBG_INIT_DEV,
				     "%s: probe: PCI device 0x%04X:0x%04X supported",
				     pcps_driver_name, vend_id, dev_id );

			return 1;
		}
	}

	_mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: probe: PCI device 0x%04X:0x%04X not supported",
		     pcps_driver_name, vend_id, dev_id );
	return 0;
}

static void
mbgclock_attach(struct device *parent, struct device *self, void *aux)
{
	int rc = 0;
	int idx;
	char devinfo[256];
	int error;
	struct mbgclock_softc *psc = device_private(self);
	struct pci_attach_args *pa = aux;
	uint16_t dev_id = PCI_PRODUCT(pa->pa_id);
	pcireg_t ecap;

	DPRINTF(DB_FOLLOW, ("mbgclock_attach(0x%p, 0x%p, 0x%p)\n", parent, self, aux));

	psc->dev = self;
	psc->pa  = *pa;
	psc->last_status = 0;
	callout_init(&psc->callout, CALLOUT_MPSAFE);

#ifdef DEBUG
	pci_conf_print(psc->pa.pa_pc, psc->pa.pa_tag, NULL);
#endif

	if ((error = pci_activate(psc->pa.pa_pc, psc->pa.pa_tag, self,
            NULL)) && error != EOPNOTSUPP) {
                aprint_error_dev(psc->dev, "cannot activate %d\n", error);
                return;
        }

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: attach for device 0x%04X",
		     pcps_driver_name, dev_id );

	psc->pddev = pcps_alloc_ddev();

	if ( psc->pddev == NULL )
	{
		_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_alloc_ddev() failed",
			     pcps_driver_name, dev_id );
		goto fail;
	}

	rc = pcps_init_ddev( psc->pddev, PCPS_BUS_PCI, dev_id );

	if ( rc != PCPS_SUCCESS )
	{
		_mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_init_ddev() failed, rc: %i",
			     pcps_driver_name, dev_id, rc );
		goto fail;
	}

	mbg_alloc_rsrcs( psc );

	rc = pcps_probe_device( psc->pddev, 0, 0 );

	if ( rc != PCPS_SUCCESS )
	{
		_mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_probe_device() failed, rc: %i",
			     pcps_driver_name, dev_id, rc );
		goto fail;
	}

	idx = snprintf(devinfo, sizeof(devinfo) - idx, "%s", _pcps_type_name( &psc->pddev->dev ) );

	if ( strlen( _pcps_sernum( &psc->pddev->dev ) ) &&
	     strcmp( _pcps_sernum( &psc->pddev->dev ), "N/A" ) )
		idx += snprintf(devinfo + idx, sizeof(devinfo) - idx, " %s", _pcps_sernum( &psc->pddev->dev ) );

	idx += snprintf(devinfo + idx, sizeof(devinfo) - idx, " (FW %X.%02X",
		       _pcps_fw_rev_num_major( _pcps_fw_rev_num( &psc->pddev->dev ) ),
		       _pcps_fw_rev_num_minor( _pcps_fw_rev_num( &psc->pddev->dev ) )
		);

	if ( _pcps_has_asic_version( &psc->pddev->dev ) )
	{
		PCI_ASIC_VERSION av;
		av = _convert_asic_version_number( psc->pddev->raw_asic_version );

		idx += snprintf(devinfo + idx, sizeof(devinfo) - idx, ", ASIC %X.%02X",
			       _pcps_asic_version_major( av ),
			       _pcps_asic_version_minor( av )
			);
	}

	idx += snprintf(devinfo + idx, sizeof(devinfo) - idx, ")" );

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: device 0x%04X attached successfully",
		     pcps_driver_name, dev_id );

	if (!pmf_device_register(self, NULL, NULL))
		aprint_error_dev(self, "couldn't establish power handler\n");

	aprint_normal(": "MBG_VENDOR" %s (rev. 0x%02x)\n", devinfo,
	    PCI_REVISION(pa->pa_class));

	mutex_init(&psc->mutex, MUTEX_DEFAULT, IPL_HIGH);

	psc->isopen = 0;

	/*
	 * prepare timecounter - to have a clean state
	 */
	psc->tc.tc_get_timecount = mbgclock_get_timecount;
	psc->tc.tc_poll_pps      = NULL;
	psc->tc.tc_counter_mask  = ~(uint)0;
	 /* actually 10 MHz as fraction counter covering 2^32 ticks */
	psc->tc.tc_frequency     = 0x100000000ULL;
	psc->tc.tc_name          = NULL; /* NULL -> not enabled */
	psc->tc.tc_quality       = 0;    /* don't override any std time counters */
	psc->tc.tc_priv          = psc;
	psc->tc.tc_next          = NULL;

	/* if we have a counter source register the timecounter */
	if (_pcps_ddev_has_fast_hr_timestamp( psc->pddev )) {
		psc->tc.tc_name = device_xname( psc->dev );

		tc_init(&psc->tc);
	}

	return;

fail:
	pci_devinfo(pa->pa_id, pa->pa_class, 0, devinfo, sizeof(devinfo));

	aprint_normal(": %s (rev. 0x%02x) - not supported\n", devinfo,
	    PCI_REVISION(pa->pa_class));

	mbg_dealloc_rsrcs( psc );

	if ( psc->pddev != NULL )
		pcps_free_ddev( psc->pddev );

	psc->pddev = NULL;

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: failed to attach device 0x%04X",
		     pcps_driver_name, dev_id );
	return;
}

static int
mbgclock_detach(struct device *self, int flags)
{
	struct mbgclock_softc *psc = device_private(self);
	PCPS_DDEV             *pddev;
	uint16_t dev_id;

	DPRINTF(DB_FOLLOW, ("mbgclock_detach(0x%p, 0x%x)\n", self, flags));

	if (psc == NULL)
		return ENXIO;

	(void)callout_halt(&psc->callout, &psc->mutex);

	if (psc->tc.tc_name != NULL) {
		tc_detach(&psc->tc);
		psc->tc.tc_name = NULL;
	}

	pddev = psc->pddev;

	dev_id  = PCI_PRODUCT(psc->pa.pa_id);

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: detach for device 0x%04X",
               pcps_driver_name, dev_id );

	/* XXX insure not outstanding requests */
	mbg_dealloc_rsrcs( psc );

	if ( psc->pddev != NULL ) {
		pcps_free_ddev( psc->pddev );
		pmf_device_deregister(self);
	}

	psc->pddev = NULL;

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: device 0x%04X detached",
               pcps_driver_name, dev_id );

	callout_destroy(&psc->callout);

	mutex_destroy(&psc->mutex);

	return 0;
}

/*------- driver -------*/

static
int mbgclockopen(dev_t dev, int flags, int mode, struct lwp *l)
{
	struct mbgclock_softc *psc = device_lookup_private(&mbgclock_cd, MBG_UNIT(dev));
	PCPS_DDEV             *pddev;

	_mbgddmsg_1( MBG_DBG_INIT_DEV, "%s: open called", pcps_driver_name );
	DPRINTF(DB_FOLLOW, ("mbgclockopen(0x%"PRIx64", 0x%x, 0x%x, %p)\n", dev, flags, mode, l));

	if (psc == NULL)
		return ENXIO;

	pddev = psc->pddev;

	if (pddev == NULL)
		return ENXIO;

	mutex_enter(&psc->mutex);
	if (!psc->isopen) {
		DPRINTF(DB_FOLLOW, ("mbgclockopen(0x%"PRIx64", 0x%x, 0x%x, %p): first open\n", dev, flags, mode, l));
		psc->isopen = 1;

		INC_REF;

		/* Clear PPS capture state on first open. */
		mutex_spin_enter(&timecounter_lock);
		memset(&psc->pps_state, 0, sizeof(psc->pps_state));
		psc->pps_state.ppscap = PPS_CAPTUREASSERT | PPS_CAPTURECLEAR;
		pps_init(&psc->pps_state);
		mutex_spin_exit(&timecounter_lock);

		callout_reset(&psc->callout, hz, mbgclock_tick, psc);
	}
	mutex_exit(&psc->mutex);

	return 0;
}

static
int mbgclockclose(dev_t dev, int flags, int mode, struct lwp *l)
{
	struct mbgclock_softc *psc = device_lookup_private(&mbgclock_cd, MBG_UNIT(dev));
	PCPS_DDEV             *pddev;

	_mbgddmsg_1( MBG_DBG_INIT_DEV, "%s: close called", pcps_driver_name );
	DPRINTF(DB_FOLLOW,("mbgclockclose(0x%"PRIx64", 0x%x, 0x%x, %p)\n", dev, flags, mode, l));

	if (psc == NULL)
		return ENXIO;

	pddev = psc->pddev;

	if (pddev == NULL)
		return ENXIO;

	mutex_enter(&psc->mutex);
	psc->isopen = 0;
	DEC_REF;
	callout_stop(&psc->callout);
	mutex_exit(&psc->mutex);

	return 0;
}

static
int mbgclockread(dev_t dev, struct uio *uio, int flags)
{
	struct mbgclock_softc *psc = device_lookup_private(&mbgclock_cd, MBG_UNIT(dev));
	PCPS_DDEV             *pddev;

	_mbgddmsg_1( MBG_DBG_INIT_DEV, "%s: read called", pcps_driver_name );
	DPRINTF(DB_FOLLOW,("mbgclockread(0x%"PRIx64", %p)\n", dev, uio));

	if (psc == NULL)
		return ENXIO;

	pddev = psc->pddev;

	if (pddev == NULL)
		return ENXIO;

	return 0;
}

static
int mbgclockwrite(dev_t dev, struct uio *uio, int flags)
{
	struct mbgclock_softc *psc = device_lookup_private(&mbgclock_cd, MBG_UNIT(dev));
	PCPS_DDEV             *pddev;

	_mbgddmsg_1( MBG_DBG_INIT_DEV, "%s: write called", pcps_driver_name );
	DPRINTF(DB_FOLLOW, ("mbgclockwrite(0x%"PRIx64", %p)\n", dev, uio));

	if (psc == NULL)
		return ENXIO;

	pddev = psc->pddev;

	if (pddev == NULL)
		return ENXIO;

	return 0;
}

#include <sys/ioctl.h>

static
int mbgclockioctl(dev_t dev, u_long cmd, void *data, int flag, struct lwp *l)
{
	struct mbgclock_softc *psc = device_lookup_private(&mbgclock_cd, MBG_UNIT(dev));
	PCPS_DDEV             *pddev;
	int rc;

	_mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: ioctl called, flag=0x%x", pcps_driver_name, flag );
	DPRINTF(DB_FOLLOW, ("mbgclockioctl(0x%"PRIx64", 0x%lx, %p, 0x%x, %p)\n",
			    dev, cmd, data, flag, l->l_proc));

	if (psc == NULL)
		return ENXIO;

	pddev = psc->pddev;

	if (pddev == NULL)
		return ENXIO;

	switch ( cmd )
	{
	case PPS_IOC_CREATE:	/* handle PPSAPI */
	case PPS_IOC_DESTROY:
	case PPS_IOC_GETPARAMS:
	case PPS_IOC_SETPARAMS:
	case PPS_IOC_GETCAP:
	case PPS_IOC_FETCH:
#ifdef PPS_SYNC
	case PPS_IOC_KCBIND:
#endif
		mutex_spin_enter(&timecounter_lock);

		rc = pps_ioctl(cmd, data, &psc->pps_state);

		if ((rc == 0) &&
		    (psc->pps_state.kcmode & PPS_CAPTUREBOTH) &&
		    (!_pcps_ddev_has_hr_time( psc->pddev ))) {
			psc->pps_state.kcmode = 0;
			mutex_spin_exit(&timecounter_lock);

			aprint_error_dev(psc->dev,
					 "ioctl(KCBIND): high resolution time stamping not supported\n");
			rc = EOPNOTSUPP;
		} else {
			mutex_spin_exit(&timecounter_lock);
		}

		break;

	default:		/* handle clock ioctls */
	{
		/* Find out which privilege level is required to execute this IOCTL command. */
		int priv_lvl = ioctl_get_required_privilege( cmd );

		/* Check if the calling process has the required privilege. */
		switch ( priv_lvl )
		{
		case MBG_REQ_PRIVL_NONE:
			/* Always allow. */
			break;

		case MBG_REQ_PRIVL_EXT_STATUS:
			/* This may require some privilege for the calling process. */
			if ((flag & FREAD) == 0 ||
			    kauth_authorize_device_passthru(l->l_cred,
							    dev,
							    KAUTH_REQ_DEVICE_RAWIO_PASSTHRU_READ,
							    &cmd))
			{
				_mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: READ/PASSTHRU READ permission denied, dev %s_%s",
					     pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				return EPERM;
			}
			break;

		case MBG_REQ_PRIVL_CFG_READ:
			/* This may require some privilege for the calling process. */
			if ((flag & FREAD) == 0 ||
			    kauth_authorize_device_passthru(l->l_cred,
							    dev,
							    KAUTH_REQ_DEVICE_RAWIO_PASSTHRU_READCONF,
							    &cmd))
			{
				_mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: READ/PASSTHRU_READCONF permission denied, dev %s_%s",
					     pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				return EPERM;
			}
			break;

		case MBG_REQ_PRIVL_SYSTEM:
			if ((flag & FWRITE) == 0 ||
			    kauth_authorize_device_passthru(l->l_cred,
							    dev,
							    KAUTH_REQ_DEVICE_RAWIO_PASSTHRU_WRITE,
							    &cmd))
			{
				_mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: WRITE/PASSTHRU_WRITE permission denied, dev %s_%s",
					     pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				return EPERM;
			}
			break;

		case MBG_REQ_PRIVL_CFG_WRITE:
			if ((flag & FWRITE) == 0 ||
			    kauth_authorize_device_passthru(l->l_cred,
							    dev,
							    KAUTH_REQ_DEVICE_RAWIO_PASSTHRU_WRITECONF,
							    &cmd))
			{
				_mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: WRITE/PASSTHRU_WRITECONF permission denied, dev %s_%s",
					     pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				return EPERM;
			}
			break;

		default:
			_mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: unknown, permission denied, dev %s_%s",
				     pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
			return EPERM;
		}

		/*
		 * now do the actual work - access the device
		 */
		rc = ioctl_switch( pddev, cmd, (void *) data, (void *) data );

		/* On success we return quickly.*/

		if ( rc != MBG_SUCCESS )
		{
			/* An error has occurred. */
			/* Generate an appropriate debug/error message */
			/* and return an error status. */

			switch ( rc )
			{
			case MBG_ERR_INV_DEV_REQUEST:
				_mbgddmsg_7( MBG_DBG_WARN, "%s: %d,%d ioctl 0x%04lX: invalid cmd %04lX, dev %s_%s",
					     pcps_driver_name, major(dev), minor(dev), cmd, IOCBASECMD( cmd ),
					     _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				rc = EINVAL;
				break;


			case MBG_ERR_NOT_SUPP_BY_DEV:
				_mbgddmsg_6( MBG_DBG_WARN, "%s: %d,%d ioctl 0x%02lX: not supported by dev %s_%s",
					     pcps_driver_name, major(dev), minor(dev), cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				rc = ENODEV;
				break;


			case MBG_ERR_NO_MEM:
				_mbgddmsg_6( MBG_DBG_WARN, "%s: %d,%d ioctl 0x%02lX: unable to allocate buffer for dev %s_%s",
					     pcps_driver_name, major(dev), minor(dev), cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				rc = ENOMEM;
				break;


			case MBG_ERR_IRQ_UNSAFE:
				_mbgddmsg_6( MBG_DBG_DETAIL, "%s: %d,%d ioctl 0x%02lX: busy since unsafe IRQ enabled, dev %s_%s",
					     pcps_driver_name, major(dev), minor(dev), cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				rc = EBUSY;
				break;


			default:  /* any access error code returned by the low level routine */
				/* or copying from or to user space */
				_mbgddmsg_7( MBG_DBG_WARN, "%s: %d,%d ioctl 0x%02lX: error %i accessing dev %s_%s",
					     pcps_driver_name, major(dev), minor(dev), cmd, rc, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
				rc = EIO;
			}
		} else {
			_mbgddmsg_6( MBG_DBG_INFO, "%s: %d,%d IOCTL 0x%02lX: success, dev %p_%s",
				     pcps_driver_name, major(dev), minor(dev), cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
			rc = 0;
		}
	}
	}

	DPRINTF(DB_FOLLOW, ("mbgclockioctl(0x%"PRIx64", 0x%lx, %p, 0x%x, %p) return rc=\n",
			    dev, cmd, data, flag, l->l_proc, rc));

	return rc;
}

static uint
mbgclock_get_timecount(struct timecounter *tc)
{
	struct mbgclock_softc *psc = tc->tc_priv;
	PCPS_TIME_STAMP ref_ts;
	int s;

	/* pick up reference counter */
        /*
	 * XXX for now this counter is NOT free running and
	 * may be re-set at any time - so this is only proof
	 * of concept code until Meinberg devices provide
	 * a truely free running counter hanging off the
	 * frequency controlled oscillator
	 * also we should only pick up the fraction part saving
	 * a PCIe bus transaction and change the coordination
	 * scheme to a generation number base one
	 */
	s = splhigh();
	/*
	 * XXX splhigh() shouldn't be necessary,
	 * but ci_curspl seems to be high after following call
	 */
	/*
	 * we only need the fraction
	 * 	do_get_fast_hr_timestamp_fraction_safe( psc->pddev, &ref_ts );
	 */
	do_get_fast_hr_timestamp_safe( psc->pddev, &ref_ts );
	splx(s);

	/*
	 * format is correct as it is a fraction counter thus
	 * our 'frequency' is nominally 0x100000000 Hz
	 */
	return ref_ts.frac;
}

static void
mbgclock_tick(void *arg)
{
	struct mbgclock_softc *psc = arg;
	struct timeval tv;

	/*
	 * TODO: should trigger poll machine
	 */

#ifdef _HAVE_PPS_REF_EVENT
	/* start work only of we need to provide PPS simulation */
	if (psc->pps_state.ppsparam.mode & PPS_CAPTUREBOTH) {
		do {
			int rc;
			struct pps_state *pps = &psc->pps_state;
			struct bintime bintime_ref_ts;
			PCPS_HR_TIME ht;
			PCPS_TIME_STAMP ref_ts;
			int valid;

			if (!_pcps_ddev_has_fast_hr_timestamp( psc->pddev )) {
				/*
				 * pick up status and PPS simulated time stamps
				 * from HR_TIME operation
				 */
				mutex_spin_enter(&timecounter_lock);

				/* pick up first system timestamp */
				pps_capture(&psc->pps_state); /* picks up capcount + th */

				_pcps_sem_inc( psc->pddev );
				rc = _pcps_read_var( psc->pddev, PCPS_GIVE_HR_TIME, ht );
				_pcps_sem_dec( psc->pddev );

				valid = (rc == MBG_SUCCESS) &&
					!((ht.status & (PCPS_INVT|PCPS_IFTM|PCPS_FREER) != 0) ||
					  ((ht.status & PCPS_SYNCD) == 0));

				/* process reftime (picks up second system timestamp) */
				if (valid) {
					bintime_ref_ts.sec  = ref_ts.sec;
					bintime_ref_ts.frac = (uint64_t)ref_ts.frac << 32;

					pps_ref_event(&psc->pps_state,
						      (psc->pps_state.ppsparam.mode & PPS_CAPTUREASSERT) ?
						      PPS_CAPTUREASSERT : PPS_CAPTURECLEAR,
						      &bintime_ref_ts,
						      PPS_REFEVNT_CAPCUR);
				}

				mutex_spin_exit(&timecounter_lock);
			} else {
				/* pick up fast time stamp when clock is ok */

				_pcps_sem_inc( psc->pddev );
				rc = _pcps_read_var( psc->pddev, PCPS_GIVE_HR_TIME, ht );
				_pcps_sem_dec( psc->pddev );

				valid = (rc == MBG_SUCCESS) &&
					!((ht.status & (PCPS_INVT|PCPS_IFTM|PCPS_FREER) != 0) ||
					  ((ht.status & PCPS_SYNCD) == 0));

				if (valid) {
					mutex_spin_enter(&timecounter_lock);

					/* pick up first system timestamp */
					pps_capture(&psc->pps_state); /* picks up capcount + th */

					/* pick up reference time */
					do_get_fast_hr_timestamp_safe( psc->pddev, &ref_ts );

					bintime_ref_ts.sec  = ref_ts.sec;
					bintime_ref_ts.frac = (uint64_t)ref_ts.frac << 32;

					/* process reftime (picks up second system timestamp) */
					pps_ref_event(&psc->pps_state,
						      (psc->pps_state.ppsparam.mode & PPS_CAPTUREASSERT) ?
						      PPS_CAPTUREASSERT : PPS_CAPTURECLEAR,
						      &bintime_ref_ts,
						      PPS_REFEVNT_CAPCUR);

					mutex_spin_exit(&timecounter_lock);
				}
			}

			/* leave when we cannot the the timestamp */
			if ( rc != MBG_SUCCESS ) {
				aprint_error_dev(psc->dev,
						 "clock read failed\n");
				break;
			}

			/* document status changes */
			if (!valid) {
				if (ht.status != psc->last_status) {
					char b[80];
					snprintb(b, sizeof b, MBG_PCPS_FMT_STATUS, ht.status);

					aprint_error_dev(psc->dev,
							 "clock not synchronized: Status=%s\n",
							 b);
					/* XXX kill timecounter - only needed while counter is linked to time */
					if (psc->tc.tc_name != NULL) {
						tc_gonebad(&psc->tc);
					}
				}
				psc->last_status = ht.status;

				break;
			}

			if ((ht.status ^ psc->last_status) &
			    (PCPS_INVT|PCPS_IFTM|PCPS_FREER|PCPS_SYNCD)) {
				char b[80];

				snprintb(b, sizeof b, MBG_PCPS_FMT_STATUS, ht.status);
				aprint_error_dev(psc->dev,
						 "clock synchronized: Status=%s\n",
						 b);
			}

			psc->last_status = ht.status;
		} while (0);		/* for break instead of goto */
	}
#endif

	/*
	 * calculate next callout shortly after next second
	 */
	microtime(&tv);
	callout_schedule(&psc->callout, 1 + hz - mstohz(tv.tv_usec / 1000));
}
