
/**************************************************************************
 *
 *  $Id: nanotime.c 1.2 2017/06/12 13:54:04 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Meinberg Library module providing functions to handle NANO_TIME
 *    and NANO_TIME_64.
 *
 * -----------------------------------------------------------------------
 *  $Log: nanotime.c $
 *  Revision 1.2  2017/06/12 13:54:04  martin
 *  Fixed build under Windows.
 *  Revision 1.1  2017/06/12 08:49:09Z  martin
 *  Initial revision.
 *
 **************************************************************************/

#define _NANOTIME
 #include <nanotime.h>
#undef _NANOTIME

#include <mbgtime.h>
#include <str_util.h>

#include <stdlib.h>


// At least MSVC++ 9.0 / Visual Studio 2008 or older
// doesn't provide strtoll()
#if defined( _MSC_VER ) && ( _MSC_VER <= 1500 )
  #define strtoll _strtoi64
#endif


/*HDR*/
/**
 * @brief Convert a ::NANO_TIME time to double
 *
 * @param[in]  p  Address of a ::NANO_TIME structure to be converted
 *
 * @return The computed number of seconds with fractions, as double
 *
 * @see ::double_to_nano_time
 */
double nano_time_to_double( const NANO_TIME *p )
{
  double d = p->secs;

  d += ( (double) p->nano_secs ) / 1e9;

  return d;

}  // nano_time_to_double



/*HDR*/
/**
 * @brief Setup a ::NANO_TIME structure from a time provided as double
 *
 * @param[out]  p  Address of a ::NANO_TIME structure to be set up
 * @param[in]   d  The time to be converted, in seconds with fractions, as double
 *
 * @see ::nano_time_to_double
 */
void double_to_nano_time( NANO_TIME *p, double d )
{
  p->secs = (int32_t) d;

  d -= (double) p->secs;

  p->nano_secs = (int32_t) ( d * 1e9 );

}  // double_to_nano_time



/*HDR*/
/**
 * @brief Print nano time into string buffer
 *
 * @param[out] s        The string buffer to be filled
 * @param[in]  max_len  Size of the output buffer for 0-terminated string
 * @param[in]  nt       The ::NANO_TIME to be printed
 */
size_t snprint_nano_time( char *s, size_t max_len, const NANO_TIME *nt )
{
  size_t n = snprintf_safe( s, max_len, "%c%li.%09li",
                            _nano_time_negative( nt ) ? '-' : '+',
                            labs( (long) nt->secs ),
                            labs( (long) nt->nano_secs ) );
  return n;

}  // snprint_nano_time



#if !defined( MBG_TGT_MISSING_64_BIT_TYPES )

/*HDR*/
/**
 * @brief Normalize a ::NANO_TIME_64 time
 *
 * After normalization, the following can be assumed:<br>
 * - nano_secs is in the range [-10^9 + 1, 10^9 - 1]<br>
 * - if secs is not 0, secs and nano_secs have the same sign
 *
 * @param[in,out] nt  The ::NANO_TIME_64 to be normalized
 */
void normalize_nano_time_64( NANO_TIME_64 *nt )
{
  int64_t additional_secs;

  // Step 1: Make sure nano seconds are in the interval [-10^9 + 1, 10^9 - 1]
  additional_secs = nt->nano_secs / NSECS_PER_SEC;
  nt->nano_secs -= additional_secs * NSECS_PER_SEC;
  nt->secs += additional_secs;

  // Step 2: Make sure seconds and nanoseconds have same sign if seconds is not 0
  if ( nt->secs > 0 && nt->nano_secs < 0 )
  {
    nt->secs -= 1;
    nt->nano_secs += NSECS_PER_SEC;
  }
  else if ( nt->secs < 0 && nt->nano_secs > 0 )
  {
    nt->secs += 1;
    nt->nano_secs -= NSECS_PER_SEC;
  }

}  // normalize_nano_time_64



/*HDR*/
/**
 * @brief Convert a ::NANO_TIME_64 time to double
 *
 * @param[in]  p  Address of a ::NANO_TIME_64 structure to be converted
 *
 * @return The computed number of seconds with fractions, as double
 *
 * @see ::double_to_nano_time_64
 */
double nano_time_64_to_double( const NANO_TIME_64 *p )
{
  double d = (double) p->secs;

  d += ( (double) p->nano_secs ) / 1e9;

  return d;

}  // nano_time_64_to_double



/*HDR*/
/**
 * @brief Setup a ::NANO_TIME_64 structure from a time as double
 *
 * @param[out]  p  Address of a ::NANO_TIME_64 structure to be set up
 * @param[in]   d  The time to be converted, in seconds with fractions, as double
 *
 * @see ::nano_time_64_to_double
 */
void double_to_nano_time_64( NANO_TIME_64 *p, double d )
{
  p->secs = (int32_t) d;

  d -= (double) p->secs;

  p->nano_secs = (int32_t) ( d * 1e9 );

}  // double_to_nano_time_64



/*HDR*/
/**
 * @brief Print a normalized ::NANO_TIME_64 into a string buffer
 *
 * @param[out] s        The string buffer to be filled
 * @param[in]  max_len  Size of the output buffer for 0-terminated string
 * @param[in]  nt       The ::NANO_TIME_64 to be printed
 */
size_t snprint_nano_time_64( char *s, size_t max_len, const NANO_TIME_64 *nt )
{
  size_t n = snprintf_safe( s, max_len, "%c%" PRId64 ".%09" PRId64,
                            _nano_time_negative( nt ) ? '-' : '+',
                            _abs64( nt->secs ),
                            _abs64( nt->nano_secs ) );
  return n;

}  // snprint_nano_time_64



static __mbg_inline /*HDR*/
/**
 * @brief Generic function to set up a ::NANO_TIME_64 structure from a string
 *
 * @param[in]   s            A string with a time in seconds, with fractions separated by decimal point
 * @param[out]  p            Address of a ::NANO_TIME_64 structure to be set up
 * @param[in]   base         The base of the conversion, i.e. 10 for a string with decimal numbers, or 16 for a hex string
 * @param[in]   frac_digits  The number of fractions required to represent 1 nanosecond, e.g.
 *                           9 for a decimal string like 0.000000001, or
 *                           TODO for a hex string like 0x00000000.00000000
 */
void str_to_nano_time_64( const char *s, NANO_TIME_64 *p, int base, int frac_digits )
{
  char *cp = NULL;

  p->secs = strtoll( s, &cp, base );

  if ( *cp == '.' )  // fractions may follow
  {
    char *ep = NULL;

    cp++;  // skip '.'

    p->nano_secs = strtoll( cp, &ep, base );

    if ( p->nano_secs )  // only if fractions not 0
    {
      int l = ep - cp;

      while ( l < frac_digits )
      {
        p->nano_secs *= base;
        l++;
      }

      while ( l > frac_digits )
      {
        p->nano_secs /= base;
        l--;
      }

      if ( p->secs < 0 )
        p->nano_secs = -p->nano_secs;
    }
  }

}  // str_to_nano_time_64



static __mbg_inline /*HDR*/
/**
 * @brief Generic function to divide ::NANO_TIME_64 value
 *
 * @param[in,out]  p    Address of a ::NANO_TIME_64 structure to be divided
 * @param[in]      div  The number by which to divide
 * @param[in]      mul  The complementary number to div, i.e. mul = 1000000000 / div
 */
void div_nano_time_64( NANO_TIME_64 *p, long div, long mul )
{
  // At least MSVC++ 9.0 / Visual Studio 2008 or older
  // doesn't provide lldiv_t and lldiv().
  #if defined( _MSC_VER ) && ( _MSC_VER <= 1500 )  //### TODO: Test this code
    int64_t rem = p->secs % div;
    p->secs /= div;
    p->nano_secs = ( rem * mul ) + ( p->nano_secs / div );
  #else
    lldiv_t lldt = lldiv( p->secs, div );
    p->secs = lldt.quot;
    p->nano_secs = ( lldt.rem * mul ) + ( p->nano_secs / div );
  #endif

}  // div_nano_time_64



/*HDR*/
/**
 * @brief Set up a ::NANO_TIME_64 structure from a string with a time in seconds and fractions
 *
 * @param[in]   s  A string with a time in seconds, with fractions separated by decimal point
 * @param[out]  p  Address of a ::NANO_TIME_64 structure to be set up
 */
void str_s_to_nano_time_64( const char *s, NANO_TIME_64 *p )
{
  str_to_nano_time_64( s, p, 10, 9 );

}  // str_s_to_nano_time_64



/*HDR*/
/**
 * @brief Set up a ::NANO_TIME_64 structure from a string with a time in milliseconds and fractions
 *
 * @param[in]   s  A string with a time in milliseconds, with fractions separated by decimal point
 * @param[out]  p  Address of a ::NANO_TIME_64 structure to be set up
 */
void str_ms_to_nano_time_64( const char *s, NANO_TIME_64 *p )
{
  // First do the same conversion as for a string representing seconds
  str_s_to_nano_time_64( s, p );

  // Now we have the value in milliseconds
  // Divide by 1000 to get seconds
  div_nano_time_64( p, 1000, 1000000UL );

}  // str_ms_to_nano_time_64

#endif  // !defined( MBG_TGT_MISSING_64_BIT_TYPES )


